import time
from datetime import date
from datetime import datetime
from datetime import timedelta
from datetime import timezone


def get_current_time_seconds() -> int:
    """Returns current unix timestamp in seconds unit

    Returns
    -------
    int
    """
    return int(time.time())


def get_current_time_miliseconds() -> int:
    """Returns current unix timestamp in milliseconds

    Returns
    -------
    int
    """
    return int(time.time_ns() // 1000000)


def get_current_time_nanoseconds() -> int:
    """Returns current unix timestamp in nanoseconds

    Returns
    -------
    int
    """
    return int(time.time_ns())


def get_current_time_miliseconds_as_string() -> str:
    """Returns current unix timestamp in milliseconds as string.

    Returns
    -------
    str
    """
    return f"{get_current_time_miliseconds()}"


def get_day_timestamps_nanoseconds(
    days_before: int, reference_date: date | None = None
) -> tuple[int, int]:
    """Returns unix start timestamp and end timestamp of given
    day in nano seconds"""
    if reference_date is None:
        reference_date = datetime.now(timezone.utc)

    # Calculate the target date by subtracting the specified number of days
    target_date = reference_date - timedelta(days=days_before)

    # Define the start and end of the target day in UTC
    start_of_day = datetime(
        target_date.year,
        target_date.month,
        target_date.day,
        0,
        0,
        0,
        tzinfo=timezone.utc,
    )
    end_of_day = start_of_day + timedelta(days=1) - timedelta(seconds=1)

    # Return the timestamps
    return int(start_of_day.timestamp() * 1e9), int(
        end_of_day.timestamp() * 1e9
    )


def get_day_timestamps_seconds(
    days_before: int, reference_date: date | None = None
) -> tuple[int, int]:
    start_of_day, end_of_day = get_day_timestamps_nanoseconds(
        days_before=days_before, reference_date=reference_date
    )
    return int(start_of_day // 1e9), int(end_of_day // 1e9)



from datetime import datetime as dt

from sqlalchemy import BigInteger
from sqlalchemy import DateTime
from sqlalchemy import Enum
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint

from xarizmi.db.models.base import Base
from xarizmi.enums import IntervalTypeEnum

from .symbol import Symbol


class CandleStick(Base):  # type: ignore
    __tablename__ = "xarizmi_candlestick"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)

    symbol_id: Mapped[int] = mapped_column(
        Integer, ForeignKey(Symbol.id), nullable=False
    )
    symbol: Mapped[Symbol] = relationship(
        "Symbol", back_populates="candlesticks"
    )

    close: Mapped[float] = mapped_column(Float, nullable=False)
    open: Mapped[float] = mapped_column(Float, nullable=False)
    high: Mapped[float] = mapped_column(Float, nullable=False)
    low: Mapped[float] = mapped_column(Float, nullable=False)
    volume: Mapped[float] = mapped_column(Float, nullable=False)
    amount: Mapped[float] = mapped_column(Float, nullable=True)
    interval_type: Mapped[str] = mapped_column(
        Enum(IntervalTypeEnum, name="interval_type_enum", create_type=True),
        nullable=False,
    )
    interval: Mapped[int] = mapped_column(BigInteger, nullable=True)
    datetime: Mapped[dt] = mapped_column(
        DateTime,
        nullable=False,
    )

    __table_args__ = (
        UniqueConstraint(
            "symbol_id",
            "interval_type",
            "interval",
            name="uix_symbol_interval",
        ),
    )



from datetime import datetime
from typing import Any
from typing import Optional

from sqlalchemy import select
from sqlalchemy.orm import Session
from sqlalchemy.sql import and_

from xarizmi.db.models.candlestick import CandleStick
from xarizmi.db.models.symbol import Symbol


def get_filtered_candlesticks(
    session: Session,
    symbol_name: Optional[str] = None,
    start_datetime: Optional[datetime] = None,
    end_datetime: Optional[datetime] = None,
    skip: int = 0,
    limit: int = 1000,
) -> list[dict[str, Any]]:

    # Base query
    query = select(
        CandleStick,  # All columns from CandleStick
        Symbol.name.label("symbol_name"),  # Symbol name with alias
    ).join(
        Symbol, CandleStick.symbol_id == Symbol.id
    )  # Join with Symbol

    # Add filters
    filters = []
    if symbol_name:
        filters.append(Symbol.name == symbol_name)
    if start_datetime:
        filters.append(CandleStick.datetime >= start_datetime)
    if end_datetime:
        filters.append(CandleStick.datetime <= end_datetime)

    if filters:
        query = query.where(and_(*filters))

    # Add ordering, skip, and limit for pagination
    query = (
        query.order_by(CandleStick.datetime.desc()).offset(skip).limit(limit)
    )

    # Execute the query
    result = session.execute(query)
    return [
        item._asdict() for item in result.all()
    ]  # Return the list of results



from pydantic import BaseModel

from xarizmi.models.currency import Currency
from xarizmi.models.exchange import Exchange


class Symbol(BaseModel):
    base_currency: Currency
    quote_currency: Currency
    fee_currency: Currency
    exchange: Exchange | None = None

    @classmethod
    def build(
        cls,
        base_currency: str,
        quote_currency: str,
        fee_currency: str,
        exchange: str | None = None,
    ) -> "Symbol":
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        """
        data = {
            "base_currency": {"name": base_currency},
            "quote_currency": {"name": quote_currency},
            "fee_currency": {"name": fee_currency},
        }
        if exchange is not None:
            data["exchange"] = {"name": exchange}
        return cls(**data)

    def to_string(self) -> str:
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        >>> symbol.to_string()
        'BTC-USD'
        """
        return (
            self.base_currency.to_string()
            + "-"
            + self.quote_currency.to_string()
        )

    def to_flat_csv(self) -> str:
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        >>> symbol.to_string()
        'BTC-USD'
        """
        return (
            self.base_currency.to_string()
            + "-"
            + self.quote_currency.to_string()
        )

    def to_dict(self) -> dict[str, str | None]:
        """
        Example
        -------
        >>> symbol = Symbol.build(
                base_currency="BTC",
                quote_currency="USD",
                fee_currency="USD",
                exchange="BINANCE",
            )
        >>> symbol.to_dict()
        {
            "base_currency": "BTC",
            "quote_currency": "USD",
            "fee_currency": "USD"
            "exchange": "BINANCE",
        }
        """
        return {
            "base_currency": self.base_currency.to_string(),
            "quote_currency": self.quote_currency.to_string(),
            "fee_currency": self.fee_currency.to_string(),
            "exchange": (
                self.exchange.to_string()
                if (self.exchange is not None)
                else None
            ),
        }



from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import relationship

from xarizmi.db.models.base import Base


class Exchange(Base):  # type: ignore
    __tablename__ = "xarizmi_exchange"
    name = Column(String, primary_key=True, unique=True)

    symbols: Mapped[list["Symbol"]] = relationship(  # type: ignore  # noqa: F821,E501
        "Symbol", back_populates="exchange"
    )



from datetime import datetime as dt

import pytz

from xarizmi.candlestick import Candlestick
from xarizmi.config import get_config
from xarizmi.db import run_db_migration
from xarizmi.db.actions.candlestick import get_filtered_candlesticks
from xarizmi.db.actions.candlestick import upsert_candlestick
from xarizmi.db.actions.exchange import bulk_upsert_exchanges
from xarizmi.db.actions.symbol import bulk_upsert_symbols
from xarizmi.db.actions.symbol import get_symbol
from xarizmi.db.client import session_scope
from xarizmi.enums import IntervalTypeEnum
from xarizmi.models.exchange import Exchange
from xarizmi.models.symbol import Symbol


def xarizmi_db_example() -> None:
    config = get_config()
    config.DATABASE_URL = "postgresql://postgres:1@localhost/xarizmi"

    run_db_migration()
    # insert exchanges to exchange table
    with session_scope() as session:
        exchanges = bulk_upsert_exchanges(
            exchanges=[
                Exchange(name=name)
                for name in [
                    "KUCOIN",
                    "crypto.com",
                    "BINANCE",
                    "COINBASE",
                ]
            ],
            session=session,
        )
        print(
            "Exchanges created in db:\n "
            f"{[vars(exchange) for exchange in exchanges]}",
            end="\n-----------\n",
        )

    # insert symbols to Symbol table
    with session_scope() as session:
        bulk_upsert_symbols(
            [
                Symbol.build(
                    base_currency="BTC",
                    quote_currency="USDT",
                    fee_currency="USDT",
                    exchange="COINBASE",
                ),
                Symbol.build(
                    base_currency="ETH",
                    quote_currency="USDT",
                    fee_currency="USDT",
                    exchange="BINANCE",
                ),
                Symbol.build(
                    base_currency="CRO",
                    quote_currency="USDT",
                    fee_currency="USDT",
                    exchange="KUCOIN",
                ),
                Symbol.build(
                    base_currency="CRO",
                    quote_currency="USD",
                    fee_currency="USD",
                    exchange="crypto.com",
                ),
            ],
            session=session,
        )

    target_symbol = Symbol.build(
        base_currency="CRO",
        quote_currency="USD",
        fee_currency="USD",
        exchange="crypto.com",
    )

    symbol_id = get_symbol(target_symbol, session=session).id
    upsert_candlestick(
        symbol_id=symbol_id,
        session=session,
        candlestick=Candlestick(
            symbol=target_symbol,
            close=1000,
            open=4,
            low=3,
            high=30000,
            volume=10000,
            amount=1,
            interval=1732385697000,
            datetime=dt(2024, 11, 23, 18, 14, 0, tzinfo=pytz.UTC),
            interval_type=IntervalTypeEnum.HOUR_1,
        ),
    )

    result = get_filtered_candlesticks(session=session, symbol_name="CRO-USD")
    print(">>>>>>>>>>>>>>>>>>>>>")
    print(result)



import pytest
from pydantic import ValidationError

from xarizmi.candlestick import Candlestick
from xarizmi.candlestick import CandlestickChart
from xarizmi.config import get_config


class TestCandlestick:
    def test(self) -> None:
        data = {
            "close": 2.5,
            "open": 1,
            "low": 0.5,
            "high": 3,
            "volume": 100,
            "amount": 150,
        }
        candle = Candlestick(**data)
        assert candle.close == 2.5
        assert candle.open == 1
        assert candle.low == 0.5
        assert candle.high == 3
        assert candle.range == 2.5
        assert candle.body == 1.5
        assert candle.is_bearish is False
        assert candle.is_bullish is True
        assert candle.doginess == 1 - (1.5 / 2.5)

        assert candle.model_dump() == pytest.approx(
            {
                "close": 2.5,
                "open": 1,
                "low": 0.5,
                "high": 3,
                "interval_type": None,
                "interval": None,
                "symbol": None,
                "volume": 100,
                "amount": 150,
                "datetime": None,
                "exchange": None,
            }
        )

    def test_flatten(self) -> None:
        data = {
            "close": 2.5,
            "open": 1,
            "low": 0.5,
            "high": 3,
            "volume": 100,
            "amount": 150,
            "exchange": {"name": "BINANCE"},
            "symbol": {
                "base_currency": {"name": "BTC"},
                "quote_currency": {"name": "USDT"},
                "fee_currency": {"name": "USDT"},
                "exchange": {"name": "BINANCE"},
            },
        }
        candle = Candlestick.model_validate(data)

        assert candle.flatten() == pytest.approx(
            {
                "close": 2.5,
                "open": 1,
                "low": 0.5,
                "high": 3,
                "interval_type": None,
                "interval": None,
                "volume": 100,
                "amount": 150,
                "datetime": None,
                "exchange": "BINANCE",
                "base_currency": "BTC",
                "quote_currency": "USDT",
            }
        )

    def test_intrinsic_range(self) -> None:
        zero_data = {
            "close": 0,
            "open": 0,
            "low": 0,
            "high": 0,
            "volume": 0,
            "amount": 0,
        }
        candle = Candlestick(**zero_data)
        assert candle.intrinsic_range == 0
        data = {
            "close": 2.5,
            "open": 1,
            "low": 0.5,
            "high": 3,
            "volume": 100,
            "amount": 150,
        }
        candle = Candlestick(**data)
        assert candle.intrinsic_range == 2.5 / 0.5

    def test_intrinsic_body(self) -> None:
        zero_data = {
            "close": 0,
            "open": 0,
            "low": 0,
            "high": 0,
            "volume": 0,
            "amount": 0,
        }
        candle = Candlestick(**zero_data)
        assert candle.intrinsic_range == 0
        data = {
            "close": 2.5,
            "open": 1,
            "low": 0.5,
            "high": 3,
            "volume": 100,
            "amount": 150,
        }
        candle = Candlestick(**data)
        assert candle.intrinsic_body == 1.5 / 2.5

    def test_intrinsic_upper_shadow(self) -> None:
        zero_data = {
            "close": 0,
            "open": 0,
            "low": 0,
            "high": 0,
            "volume": 0,
            "amount": 0,
        }
        candle = Candlestick(**zero_data)
        assert candle.intrinsic_upper_shadow == 0
        data = {
            "close": 2.5,
            "open": 1,
            "low": 0.5,
            "high": 3,
            "volume": 100,
            "amount": 150,
        }
        candle = Candlestick(**data)
        assert candle.upper_shadow == 0.5
        assert candle.intrinsic_upper_shadow == 0.5 / 2.5

    def test_intrinsic_lower_shadow(self) -> None:
        zero_data = {
            "close": 0,
            "open": 0,
            "low": 0,
            "high": 0,
            "volume": 0,
            "amount": 0,
        }
        candle = Candlestick(**zero_data)
        assert candle.intrinsic_lower_shadow == 0
        data = {
            "close": 2.5,
            "open": 1,
            "low": 0.5,
            "high": 3,
            "volume": 100,
            "amount": 150,
        }
        candle = Candlestick(**data)
        assert candle.lower_shadow == 0.5
        assert candle.intrinsic_lower_shadow == 0.5 / 2.5

    def test_negative_price(self) -> None:
        # Given a data with negative price
        data = {
            "close": -2.5,
            "open": 1,
            "low": 0.5,
            "high": 3,
            "volume": 100,
            "amount": 150,
        }
        # When Candlestick constructor is called
        # Then I should see ValidationError
        with pytest.raises(ValidationError):
            Candlestick(**data)

    def test_is_dogi(self) -> None:
        # Given a data with negative price
        data = {
            "close": 30,
            "open": 20,
            "low": 10,
            "high": 40,
            "volume": 100,
            "amount": 150,
        }
        # When a candlestick with this data is created
        c = Candlestick(**data)
        # Then I should have
        assert c.is_doji is False

        # And when I change the config
        config = get_config()
        config.DOJINESS_THRESHOLD = 0.1
        # When a candlestick with this data is created
        c = Candlestick(**data)
        # Then I should have
        assert c.is_doji is True
        config.reset()

    def test_is_dogi_when_range_is_zero(self) -> None:
        # Given a data with high and low are in same price
        data = {
            "close": 30,
            "open": 30,
            "low": 30,
            "high": 30,
            "volume": 100,
            "amount": 150,
        }
        # When a candlestick with this data is created
        c = Candlestick(**data)
        # Then I should have
        assert c.is_doji is False

    def test_is_doginess_when_range_is_zero(self) -> None:
        # Given a data with high and low are in same price
        data = {
            "close": 30,
            "open": 30,
            "low": 30,
            "high": 30,
            "volume": 100,
            "amount": 150,
        }
        # When a candlestick with this data is created
        c = Candlestick(**data)
        # Then I should have
        assert c.doginess == 0


class TestCandlestickChart:

    def test(self) -> None:
        data = {
            "candles": [
                {
                    "low": 0.61873,
                    "high": 0.727,
                    "close": 0.714,
                    "open": 0.71075,
                    "interval_type": "1week",
                    "interval": 604800,
                    "symbol": None,
                    "volume": 100,
                    "amount": 150,
                    "datetime": None,
                    "exchange": None,
                },
                {
                    "low": 0.65219,
                    "high": 0.75,
                    "close": 0.70238,
                    "open": 0.71075,
                    "interval_type": "1week",
                    "interval": 604800,
                    "symbol": None,
                    "volume": 100,
                    "amount": 150,
                    "datetime": None,
                    "exchange": None,
                },
                {
                    "low": 0.64801,
                    "high": 0.92,
                    "close": 0.8404,
                    "open": 0.70238,
                    "interval_type": "1week",
                    "interval": 604800,
                    "symbol": None,
                    "volume": 100,
                    "amount": 150,
                    "datetime": None,
                    "exchange": None,
                },
            ]
        }

        chart = CandlestickChart(**data)

        assert chart.model_dump() == pytest.approx(data)

    def test_to_df(self) -> None:
        data = {
            "candles": [
                {
                    "low": 0.61873,
                    "high": 0.727,
                    "close": 0.714,
                    "open": 0.71075,
                    "interval_type": "1week",
                    "interval": 604800,
                    "symbol": None,
                    "volume": 100,
                    "amount": 150,
                    "datetime": None,
                },
                {
                    "low": 0.65219,
                    "high": 0.75,
                    "close": 0.70238,
                    "open": 0.71075,
                    "interval_type": "1week",
                    "interval": 604800,
                    "symbol": None,
                    "volume": 100,
                    "amount": 150,
                    "datetime": None,
                },
                {
                    "low": 0.64801,
                    "high": 0.92,
                    "close": 0.8404,
                    "open": 0.70238,
                    "interval_type": "1week",
                    "interval": 604800,
                    "symbol": None,
                    "volume": 100,
                    "amount": 150,
                    "datetime": None,
                },
            ]
        }

        chart = CandlestickChart(**data)

        df = chart.to_df()
        assert df["open"].tolist() == [0.71075, 0.71075, 0.70238]
        assert df["close"].tolist() == [0.714, 0.70238, 0.8404]
        assert df["low"].tolist() == [0.61873, 0.65219, 0.64801]
        assert df["high"].tolist() == [0.727, 0.75, 0.92]

    def test_btc_usdt_monthly_data(
        self, btc_usdt_monthly_data: list[dict[str, int | float]]
    ) -> None:
        # Given dataset of BTC-USDT candlestick data
        # When CandlestickChart is called
        c = CandlestickChart.model_validate({"candles": btc_usdt_monthly_data})
        # Then I should have
        assert len(c.candles) == 80

    def test_get_local_minima_candles(
        self, btc_usdt_monthly_data: list[dict[str, int | float]]
    ) -> None:
        # Given dataset of BTC-USDT candlestick data
        # And a candlestick chart with this data
        c = CandlestickChart.model_validate({"candles": btc_usdt_monthly_data})
        # When CandlestickChart.get_local_minima_candles is called
        values = c.get_local_minimas()
        # Then I should have
        assert len(values) == 17
        assert values[:17] == [
            5005.0,
            5950.0,
            6412.000001,
            5736.812481,
            5865.0,
            3160.000001,
            6437.0,
            3800.0,
            9825.0,
            28801.8,
            32918.7,
            17614.7,
            15473.6,
            19558.1,
            24807.5,
            24901.7,
            56560.0,
        ]

        # And when CandlestickChart.get_local_minima_candles is called
        # with only_significant_digit argument set to True
        values = c.get_local_minimas(only_significant_digit=True)
        # Then I should have
        assert values == [
            5000.0,
            6000.0,
            6000.0,
            6000.0,
            6000.0,
            3000.0,
            6000.0,
            4000.0,
            10000.0,
            30000.0,
            30000.0,
            20000.0,
            20000.0,
            20000.0,
            20000.0,
            20000.0,
            60000.0,
        ]

    def test_get_local_minima_candles_raise_error_for_invalid_argument(
        self, btc_usdt_monthly_data: list[dict[str, int | float]]
    ) -> None:
        # Given dataset of BTC-USDT candlestick data
        # And a candlestick chart with this data
        c = CandlestickChart.model_validate({"candles": btc_usdt_monthly_data})
        # When CandlestickChart.get_local_minima_candles is called with
        # wrong argument
        # Then I should have
        with pytest.raises(ValueError):
            c.get_local_minimas(price_type="DOES_NOT_EXIST")



from typing import TypedDict

from sqlalchemy.orm import Session

from xarizmi.db.client import session_scope
from xarizmi.db.mappers.base import PydanticModelDbMapper
from xarizmi.db.models.symbol import Symbol as SqlAlchemySymbol
from xarizmi.models.currency import Currency
from xarizmi.models.exchange import Exchange as PydanticExchange
from xarizmi.models.symbol import Symbol as PydanticSymbol


class SYMBOL_FLAT_TYPED_DICT(TypedDict):
    base_currency: str
    quote_currency: str
    fee_currency: str
    exchange: str


class SymbolDbMapper(PydanticModelDbMapper):
    PRIMARY_KEY_FIELD = ["id"]
    COMPUTED_FIELD_KEYS = ["name"]

    def __init__(self, obj: SqlAlchemySymbol | PydanticSymbol) -> None:
        super().__init__(
            pydantic_class=PydanticSymbol, sqlalchemy_class=SqlAlchemySymbol
        )
        self._refresh(obj)

    @property
    def pydantic_obj(self) -> PydanticSymbol:
        return self._pydantic_obj  # type: ignore

    @property
    def sqlalchemy_obj(self) -> SqlAlchemySymbol:
        return self._sqlalchemy_obj  # type: ignore

    def _to_pydantic(self, obj: SqlAlchemySymbol) -> PydanticSymbol:  # type: ignore  # noqa: E501
        return PydanticSymbol(
            base_currency=Currency(name=str(obj.base_currency)),
            quote_currency=Currency(name=str(obj.quote_currency)),
            fee_currency=Currency(name=str(obj.fee_currency)),
            exchange_name=PydanticExchange(name=str(obj.exchange_name)),
        )

    def _to_sqlalchemy(self, obj: PydanticSymbol) -> SqlAlchemySymbol:  # type: ignore  # noqa: E501
        return SqlAlchemySymbol(
            base_currency=obj.base_currency.name,
            quote_currency=obj.quote_currency.name,
            fee_currency=obj.fee_currency.name,
            exchange_name=obj.exchange.name,  # type: ignore
        )

    def _does_exist_in_session(self, session: Session) -> bool:
        record = (
            session.query(SqlAlchemySymbol)
            .filter_by(
                base_currency=self.sqlalchemy_obj.base_currency,
                quote_currency=self.sqlalchemy_obj.quote_currency,
            )
            .first()
        )
        if record is None:
            return False
        else:
            return True

    def get(self, session: None | Session = None) -> SqlAlchemySymbol | None:
        return super().get(session=session)  # type: ignore

    def _get_in_session(self, session: Session) -> SqlAlchemySymbol | None:  # type: ignore  # noqa: E501
        record = (
            session.query(SqlAlchemySymbol)
            .filter_by(
                base_currency=self.sqlalchemy_obj.base_currency,
                quote_currency=self.sqlalchemy_obj.quote_currency,
                exchange_name=self.sqlalchemy_obj.exchange_name,
            )
            .first()
        )
        return record

    def _upsert_in_session(self, session: Session) -> None:
        record = self._get_in_session(session=session)
        if record:
            for key in self.sqlalchemy_obj.__table__.columns.keys():
                if key not in (
                    SymbolDbMapper.PRIMARY_KEY_FIELD
                    + SymbolDbMapper.COMPUTED_FIELD_KEYS
                ):  # Exclude primary key and computed keys from updates
                    setattr(record, key, getattr(self.sqlalchemy_obj, key))
        else:
            session.merge(self.sqlalchemy_obj)

    def upsert(self, session: Session | None = None) -> None:
        if session is None:
            with session_scope() as session:
                self._upsert_in_session(session=session)
        else:
            self._upsert_in_session(session=session)

    @staticmethod
    def bulk_upsert_from_data(
        data_list: list[SYMBOL_FLAT_TYPED_DICT], session: Session | None = None
    ) -> None:
        """
        data_list should be like this
        [
            {
                "base_currency": "BTC",
                "quote_currency": "USDT",
                "fee_currency": "USDT",
                "exchange_name": "BINANCE",
            },
            ...
        ]
        """
        symbol_dbs: list[SymbolDbMapper] = []
        for item in data_list:
            symbol_dbs.append(SymbolDbMapper(PydanticSymbol.build(**item)))
        if session is None:
            with session_scope() as session:
                for symbol_db in symbol_dbs:
                    symbol_db.upsert(session=session)
        else:
            for symbol_db in symbol_dbs:
                symbol_db.upsert(session=session)
